/*
 * Project "fingerprintGUI": Services for fingerprint authentication on Linux
 * Module: DeviceHandler.cpp, DeviceHandler.h
 * Purpose: A module to deal with devices
 *
 * @author  Wolfgang Ullrich
 * Copyright (C) 2008-2009  Wolfgang Ullrich
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * DeviceHandler maintains a list of all available USB devies and a list of all
 * attached and supported fingerprint reader devices.
 *
 * In the first step a list of all known vendor-id and device-id is derived from
 * the file usb.ids, that is located in the programm ressources and was found
 * on http://www.linux-usb.org/usb.ids. This is only to be able to show a list of 
 * all attached usb devices at GUI startup/rescan.
 *
 * In the second step the list of attached usb devices is created and can be 
 * obtained by getAttachedUsbDevices().
 *
 * In the third step a list of attached and supported fingerprint readers is 
 * created. On adding a device the signal "deviceAdded" is emitted.
 *
 * Calling getIdentifiers returns a list of fingerprint devices that can identify.
 * Calling getVerifiers returns a list of fingerprint devices that can verify only.
 */

#include <string>
#include <iostream>
#include <fstream>
#include <usb.h>
#include <qstring.h>
#include <qcombobox.h>
#include "UsbDevice.h"
#include "FingerprintDevice.h"
#include "DeviceHandler.h"

using namespace std;

DeviceHandler::DeviceHandler(display_name_mode mode){
    knownUSBDevices=NULL;
    attachedUSBDevices=NULL;
    discoveredFpDevices=NULL;
#ifdef HAS_LIBBSAPI
    discoveredBsDevices=NULL;
#endif
    fingerprintDevices=NULL;
    currentDevice=NULL;
    identifierDevices=NULL;
    verifierDevices=NULL;
    displayNameMode=mode;
    if(mode==DISPLAY_VENDOR_NAME)   //used for GUI
        getKnownUSBDevices();
}

// public getters and setters --------------------------------------------------
USBDevice* DeviceHandler::getAttachedUsbDevices(){
    return attachedUSBDevices;
}

FingerprintDevice* DeviceHandler::getFingerprintDevices(){
    return fingerprintDevices;
}

FingerprintDevice* DeviceHandler::getCurrentDevice(){
    return currentDevice;
}

void DeviceHandler::setDisplayNameMode(display_name_mode mode){
    displayNameMode=mode;
}

// moves devices that can identify to "identifierDevices" list and
// those can not identify to "verifierDevices"
// Sets fingerprintDevices to NULL
FingerprintDevice* DeviceHandler::getIdentifiers(){
    if(fingerprintDevices==NULL){
        syslog(LOG_ERR,"No fingerprint devices found!");
        return NULL;
    }
    FingerprintDevice *current=NULL;
    FingerprintDevice *last_i=NULL;
    FingerprintDevice *last_v=NULL;
    FingerprintDevice *next=NULL;
    identifierDevices=NULL;
    verifierDevices=NULL;
    for(current=fingerprintDevices;current!=NULL;current=next){
        next=current->next;
        syslog(LOG_INFO,"Discovered device: %s.",current->getDisplayName(DISPLAY_DRIVER_NAME)->data());
        if(!current->canIdentify()){
            syslog(LOG_INFO,"Device can not identify.");
            if(verifierDevices==NULL){
                verifierDevices=current;
            }
            else{
                last_v->next=current;
            }
            last_v=current;
        }
        else{
            syslog(LOG_INFO,"Device can identify.");
            if(identifierDevices==NULL){
                identifierDevices=current;
            }
            else{
                last_i->next=current;
            }
            last_i=current;
        }
        current->next=NULL;
    }
    fingerprintDevices=NULL;
    return identifierDevices;
}

FingerprintDevice* DeviceHandler::getVerifiers(){
    if(fingerprintDevices==NULL){   //No devices or "getIdentifiers" was called already
        return verifierDevices;
    }
    getIdentifiers();   //Separate them first
    return verifierDevices;
}

// public methods --------------------------------------------------------------
void DeviceHandler::rescan(){
    //free old device lists
    for(FingerprintDevice* fp=fingerprintDevices;fp!=NULL;){
        FingerprintDevice* fp_p=fp;
        fp=fp->next;
        delete fp_p;
    }
    fingerprintDevices=NULL;

    for(FingerprintDevice* fp=identifierDevices;fp!=NULL;){
        FingerprintDevice* fp_p=fp;
        fp=fp->next;
        delete fp_p;
    }
    identifierDevices=NULL;

    for(FingerprintDevice* fp=verifierDevices;fp!=NULL;){
        FingerprintDevice* fp_p=fp;
        fp=fp->next;
        delete fp_p;
    }
    verifierDevices=NULL;

    for(USBDevice* usb=attachedUSBDevices;usb!=NULL;){
        USBDevice* usb_p=usb;
        usb=usb->next;
        delete usb_p;
    }
    attachedUSBDevices=NULL;
    
    if(findAttachedUSBDevices()>0){
        if(discoveredFpDevices!=NULL)
            fp_dscv_devs_free(discoveredFpDevices);

#ifdef HAS_LIBBSAPI
        if(discoveredBsDevices!=NULL)
            ABSFree(discoveredBsDevices);

        if(ABSEnumerateDevices("usb", &discoveredBsDevices)==ABS_STATUS_OK){
            if(discoveredBsDevices->NumDevices>0){
                for(uint i=0;i<discoveredBsDevices->NumDevices;i++){
                    syslog(LOG_INFO,"Have libbsapi-device %s.",discoveredBsDevices->List[i].DsnSubString);
                    addDevice(new FingerprintDevice(&discoveredBsDevices->List[i],knownUSBDevices));
                }
            }
        }
        else{
            syslog(LOG_ERR,"ABSEnumerateDevices() failed.");
        }
#endif
        discoveredFpDevices=fp_discover_devs();

        if(discoveredFpDevices!=NULL){
            for(int i=0;discoveredFpDevices[i]!=NULL;i++){
                addDevice(new FingerprintDevice(discoveredFpDevices[i],knownUSBDevices));
            }
        }
    }
    emit rescanFinished();
}

// slots -----------------------------------------------------------------------
USBDevice* DeviceHandler::findKnownDevice(int vendor, int device){
    string v="0000  unknown vendor";
    string d="0000  unknown device";

    USBDevice* usb=new USBDevice(v,d);
    USBDevice* walk=knownUSBDevices;
    usb->vendorID=vendor;
    usb->deviceID=device;

    while(walk!=NULL){
        if(walk->vendorID==vendor){
            usb->vendorName=walk->vendorName;
            if(walk->deviceID==device){
                usb->deviceName=walk->deviceName;
                return usb;
            }
        }
        walk=walk->next;
    }
    return usb;
}

void DeviceHandler::setCurrentDevice(int index){
    currentDevice=fingerprintDevices;
    for(int i=index;i>0&&currentDevice!=NULL;i--)
        currentDevice=currentDevice->next;
    if(currentDevice!=NULL)
        syslog(LOG_DEBUG,"Current device set to %d -- %s.",index,currentDevice->getDisplayName(DISPLAY_DRIVER_NAME)->data());
}

// private helpers -------------------------------------------------------------
void DeviceHandler::addDevice(FingerprintDevice* fpDevice){
    fpDevice->next=NULL;
    if(fingerprintDevices==NULL){
        fingerprintDevices=fpDevice;
    }
    else{
        FingerprintDevice* fpDev=fingerprintDevices;
        while(fpDev->next!=NULL){
#ifdef HAS_LIBBSAPI
            // don't add devices found by libfprint that were found by libbsapi before
            if(!fpDev->isFpDevice()){ // this is a bs device
                if(fpDev->vendorId==fpDevice->vendorId && fpDev->productId==fpDevice->productId){
                    // this device is handled by libbsapi; don't add as fp device again
                    syslog(LOG_DEBUG,"Device %s is handled by libbsapi already; not added.",fpDevice->getDisplayName(DISPLAY_DRIVER_NAME)->data());
                    delete fpDevice;
                    return;
                }
            }
#endif
            fpDev=fpDev->next;
        }
#ifdef HAS_LIBBSAPI
        // don't add devices found by libfprint that were found by libbsapi before
        if(!fpDev->isFpDevice()){ // this is a bs device
            if(fpDev->vendorId==fpDevice->vendorId && fpDev->productId==fpDevice->productId){
                // this device is handled by libbsapi; don't add as fp device again
                syslog(LOG_DEBUG,"Device %s is handled by libbsapi already; not added.",fpDevice->getDisplayName(DISPLAY_DRIVER_NAME)->data());
                delete fpDevice;
                return;
            }
        }
#endif
        fpDev->next=fpDevice;
    }
    syslog(LOG_DEBUG,"Added %s.",fpDevice->getDisplayName(DISPLAY_DRIVER_NAME)->data());
    emit deviceAdded(*fpDevice->getDisplayName(displayNameMode));
}

int DeviceHandler::findAttachedUSBDevices(){
    struct usb_bus *bus;
    struct usb_device *dev=NULL;
    USBDevice* lastDevice=NULL;

    usb_init();
    usb_find_busses();
    usb_find_devices();

    for(bus=usb_busses;bus;bus=bus->next){
        for (dev=bus->devices;dev;dev=dev->next) {
            if(dev->descriptor.idVendor!=0||dev->descriptor.idProduct!=0){  //attched device
                USBDevice* usb=findKnownDevice(dev->descriptor.idVendor,dev->descriptor.idProduct);
                usb->device=dev;
                string v=string(usb->vendorName);
                string p=string(usb->deviceName);
                syslog(LOG_DEBUG,"Found USB device: %s/%s.",v.data(),p.data());
                if(lastDevice==NULL){
                    attachedUSBDevices=usb;
                }
                else{
                    lastDevice->next=usb;
                }
                lastDevice=usb;
            }
        }
    }
    return true;
}

int DeviceHandler::getKnownUSBDevices(){
    string vendor;
    string device;
    USBDevice* lastDevice=NULL;
    QString endOfDeviceId=END_OF_DEVICE_IDS;

    QFile idFile(":/new/prefix1/res/usb.ids");
    if (!idFile.open(QIODevice::ReadOnly | QIODevice::Text))
        return false;
    
    QTextStream in(&idFile);
    while (!in.atEnd()) {
        QString line = in.readLine();
        if(line.compare(endOfDeviceId)==0)
            break;
        if(line.length()!=0&&!line.startsWith("#")){  //comment lines and empty lines
            if(line.startsWith("\t")){  //device
                device=(line.toStdString()).substr(1);
                USBDevice* usb=new USBDevice(vendor,device);
                if(lastDevice==NULL){
                    knownUSBDevices=usb;
                }
                else{
                    lastDevice->next=usb;
                }
                lastDevice=usb;
            }
            else{               //vendor
                vendor=line.toStdString();
            }
        }
    }
    idFile.close();
    return true;
}
