/*
 * Project "fingerprintGUI": Services for fingerprint authentication on Linux
 * Module: Fingerprint.cpp, Fingerprint.h
 * Purpose: Repesents a single fingerprint, it's data and the asigned device,
 * handles some GUI interaction.
 *
 * @author  Wolfgang Ullrich
 * Copyright (C) 2008-2009  Wolfgang Ullrich
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sstream>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <usb.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "Fingerprint.h"
#include "FingerprintDevice.h"
#include "DeviceHandler.h"
#include "UsbDevice.h"

Fingerprint::Fingerprint(int finger,FingerprintDevice *d,QLabel* textLabels[5],QLabel* iconLabels[5]){
    valid=false;
    busy=false;
    data=NULL;
    this->finger=finger;
    this->textLabels[0]=textLabels[0];
    this->textLabels[1]=textLabels[1];
    this->textLabels[2]=textLabels[2];
    this->textLabels[3]=textLabels[3];
    this->textLabels[4]=textLabels[4];
    this->iconLabels[0]=iconLabels[0];
    this->iconLabels[1]=iconLabels[1];
    this->iconLabels[2]=iconLabels[2];
    this->iconLabels[3]=iconLabels[3];
    this->iconLabels[4]=iconLabels[4];
    setAcquireStage(-1);
    device=d;
    device->setMode(MODE_ACQUIRE);
    if(loadData()){         //existing fingerprint found
        valid=true;
    }
}

// public getters and setters --------------------------------------------------
FingerprintDevice *Fingerprint::getDevice(){
    return device;
}

bool Fingerprint::isValid(){
    return valid;
}

void Fingerprint::modeVerify(){
    device->setMode(MODE_VERIFY);
}

bool Fingerprint::swipe(){
    if(device==NULL||busy)return false;
    busy=true;
    syslog(LOG_DEBUG,"Swipe finger at %s.",device->getDisplayName(DISPLAY_DRIVER_NAME)->data());
    device->start();
    syslog(LOG_DEBUG,"Thread started.");
    return true;
}

// slots -----------------------------------------------------------------------
void Fingerprint::newAcquireResult(int result){
    int rc=result;
    
    setAcquireStage(result);
    switch(result){
        case FP_ENROLL_COMPLETE:
            syslog(LOG_DEBUG,"Acquire complete.");
            valid=false;
            dataSize=device->getData(&data);
            if(dataSize==0){
                rc=FP_ENROLL_FAIL;
                break;
            }
            if(saveData()){
                valid=true;
            }
            else{
                rc=FP_ENROLL_FAIL;
            }
            break;
        case FP_ENROLL_FAIL:
            syslog(LOG_DEBUG,"Acquire failed.");
            valid=false;
            break;
        default:
            return;
    }
    emit acquireFinished(rc);
}

void Fingerprint::newVerifyResult(int result){
    setVerifyStage(result);
    switch(result){
        case FP_VERIFY_NO_MATCH:
            syslog(LOG_DEBUG,"Verify failed.");
            valid=false;
            break;
        case FP_VERIFY_MATCH:
            syslog(LOG_DEBUG,"Verify complete.");
            valid=true;
            break;
        default:
            return;
    }
    emit verifyFinished(result);
}

// private helpers -------------------------------------------------------------
string Fingerprint::getFilename(){    //returns filename to store the users fingerprint data to
    const char *homedir=getenv("HOME");
    std::ostringstream o;
    string fname=string(homedir);
    fname.append(DATA_DIR);
    std::ifstream check1(fname.data());
    if(!check1){                         //directory does not exists yet
        mkdir(fname.data(),S_IRWXU);
    }
    if(device->isFpDevice()){
        fname.append(FP_DATA_DIR);
    }
#ifdef HAS_LIBBSAPI
    else{
        fname.append(BS_DATA_DIR);
    }
#endif
    std::ifstream check2(fname.data());
    if(!check2){                         //directory does not exists yet
        mkdir(fname.data(),S_IRWXU);
    }
    o << finger;
    fname.append(o.str());
    fname.append(DATA_EXT);
    return fname;
}

bool Fingerprint::loadData(){   //loads fingerprint data for this finger and this device and puts it into the device
    string filename=getFilename();
    ifstream birFile(filename.data(),ios_base::binary);
    if(birFile.is_open()){
        birFile.seekg(0,ios::end);
        dataSize=birFile.tellg();
        if(dataSize==0){
            birFile.close();
            return false;
        }
        birFile.seekg(0,ios::beg);
        data=new char[dataSize];
        birFile.read((char*)data,dataSize);
        birFile.close();
        syslog(LOG_DEBUG,"Load data from %s (%d bytes).",filename.data(),dataSize);
    }
    else{
        return false;
    }
    device->setData(data,dataSize);
    return true;
}

bool Fingerprint::saveData(){  //saves fingerprint data for this finger and this device to disk
    string filename=getFilename();
    ofstream birFile(filename.data(),ios_base::binary|ios_base::out);
    if(birFile.is_open()){
        birFile.write((const char*)data,dataSize);
        syslog(LOG_DEBUG,"Save data to %s (%d bytes).",filename.data(),dataSize);
        birFile.close();
    }
    else{
        syslog(LOG_ERR,"Could not save data to disk.");
        return false;
    }
    return true;            
}

void Fingerprint::setAcquireStage(int result){
    if(result==-1){ //initial
        stage=0;
        textLabels[0]->setText("waiting...");
        textLabels[1]->setText("");
        textLabels[2]->setText("");
        textLabels[3]->setText("");
        textLabels[4]->setText("");
        iconLabels[0]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-wait.png")));
        iconLabels[1]->setPixmap(NULL);
        iconLabels[2]->setPixmap(NULL);
        iconLabels[3]->setPixmap(NULL);
        iconLabels[4]->setPixmap(NULL);
        return;
    }
    switch(result){
        case FP_ENROLL_COMPLETE:
            syslog(LOG_DEBUG,"Acquire stage %d OK.",stage);
            textLabels[stage]->setText("OK");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-ok.png")));
            break;
        case FP_ENROLL_PASS:
            syslog(LOG_DEBUG,"Acquire stage %d OK.",stage);
            textLabels[stage]->setText("OK");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-ok.png")));
            if(stage>=2)return;//to be done: Handle more than 3 stages.
            stage++;
            syslog(LOG_DEBUG,"Acquire stage %d waiting...",stage);
            textLabels[stage]->setText("waiting...");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-wait.png")));
            break;
        case FP_ENROLL_RETRY_TOO_SHORT:
            syslog(LOG_DEBUG,"Acquire stage %d swipe too short...",stage);
            textLabels[stage]->setText("Swipe too short...");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-error.png")));
            break;
        case FP_ENROLL_RETRY_CENTER_FINGER:
            syslog(LOG_DEBUG,"Acquire stage %d please center...",stage);
            textLabels[stage]->setText("Please center...");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-error.png")));
            break;
        case FP_ENROLL_RETRY:
        case FP_ENROLL_RETRY_REMOVE_FINGER:
            syslog(LOG_DEBUG,"Acquire stage %d try again...",stage);
            textLabels[stage]->setText("Try again...");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-error.png")));
            break;
        default:
            setAcquireStage(-1);
            return;
    }
}

void Fingerprint::setVerifyStage(int result){
    switch(result){
        case FP_VERIFY_NO_MATCH:
            syslog(LOG_DEBUG,"Verify stage %d no match",stage);
            textLabels[stage]->setText("No match!");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-error.png")));
            break;
        case FP_VERIFY_MATCH:
            syslog(LOG_DEBUG,"Verify stage %d OK",stage);
            textLabels[stage]->setText("OK");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-ok.png")));
            break;
        case FP_VERIFY_RETRY_TOO_SHORT:
            syslog(LOG_DEBUG,"Verify stage %d swipe too short...",stage);
            textLabels[stage]->setText("Swipe too short...");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-error.png")));
            break;
        case FP_VERIFY_RETRY_CENTER_FINGER:
            syslog(LOG_DEBUG,"Verify stage %d please center...",stage);
            textLabels[stage]->setText("Please center...");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-error.png")));
            break;
        case FP_VERIFY_RETRY:
        case FP_VERIFY_RETRY_REMOVE_FINGER:
            syslog(LOG_DEBUG,"Verify stage %d try again...",stage);
            textLabels[stage]->setText("Try again...");
            iconLabels[stage]->setPixmap(QPixmap(QString::fromUtf8(":/new/prefix1/res/fp-error.png")));
            break;
    }
}
