/*
 * Project "fingerprintGUI": Services for fingerprint authentication on Linux
 * Module: FingerprintDevice.cpp, FingerprintDevice.h
 * Purpose: An USB device to deal with a fingerprint sensor
 *
 * @author  Wolfgang Ullrich
 * Copyright (C) 2008-2009  Wolfgang Ullrich
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <pthread.h>
#include <string>
#include <usb.h>
#include <iostream>
#include "DeviceHandler.h"
#include "FingerprintDevice.h"
#include "UsbDevice.h"

using namespace std;

// Constructor for "libfprint" type devices
FingerprintDevice::FingerprintDevice(struct fp_dscv_dev *fp,USBDevice *knownUSBDevices){
    fpDevice=fp;
    isFp=true;
    fpData=NULL;
    dev=NULL;
#ifdef HAS_LIBBSAPI
    bsData=NULL;
    conn=0;
    absFlag=0;
#endif
    vendorId=fp->udev->descriptor.idVendor;
    productId=fp->udev->descriptor.idProduct;

    driverName=string("");
    displayName=string("");
    driverName.append(fp_driver_get_full_name(fp_dscv_dev_get_driver(fpDevice)));
    syslog(LOG_INFO,"initializing libfprint device (vend/prod) 0x%x/0x%x, driver: %s",vendorId,productId,driverName.data());
    if(knownUSBDevices!=NULL){
        for(USBDevice *u=knownUSBDevices;u!=NULL;u=u->next){
            if(u->vendorID==vendorId&&u->deviceID==productId){
                displayName.append(string(u->vendorName));
                displayName.append(" ");
                displayName.append(string(u->deviceName));
                break;
            }
        }
    }
    else{
        displayName.append(fp_driver_get_full_name(fp_dscv_dev_get_driver(fpDevice)));
    }
}

#ifdef HAS_LIBBSAPI
// Constructor for "libbsapi" type devices
FingerprintDevice::FingerprintDevice(struct abs_device_list_item *bs,USBDevice *knownUSBDevices){
    fpDevice=NULL;
    isFp=false;
    fpData=NULL;
    bsData=NULL;
    conn=0;
    absFlag=0;
    
    driverName=string("");
    displayName=string("");
    bsDevice=string(strdup(bs->DsnSubString));
    driverName.append("libbsapi");
     
    string s=string(strdup(bs->DsnSubString));
    if(s.find("VID_",0)==string::npos || s.find("_PID_",0)==string::npos){
        syslog(LOG_ERR,"FIXME! Couldn't parse vendor/product.");
        return;
    }
    vendorId=strtol(s.substr(s.find("VID_",0)+4,4).data(),NULL,16);
    productId=strtol(s.substr(s.find("_PID_",0)+5,4).data(),NULL,16);
    
    syslog(LOG_INFO,"initializing libbsapi device (vend/prod) 0x%x/0x%x",vendorId,productId);
    if(knownUSBDevices!=NULL){
        for(USBDevice *u=knownUSBDevices;u!=NULL;u=u->next){
            if(u->vendorID==vendorId&&u->deviceID==productId){
                displayName.append(string(u->vendorName));
                displayName.append(" ");
                displayName.append(string(u->deviceName));
                break;
            }
        }
    }
    else{
        displayName.append(strdup(bs->DsnSubString));
    }
}
#endif

// public getters and setters --------------------------------------------------
string* FingerprintDevice::getDisplayName(int mode){
    if(mode==DISPLAY_DRIVER_NAME)
        return &driverName;
    return &displayName;
}

void FingerprintDevice::setMode(int m){
    mode=m;
}

void FingerprintDevice::setIdentifyData(void **iData,int n){
    if(mode==MODE_IDENTIFY){
        identifyData=iData;
        numIdentify=n;
    }
    if(mode==MODE_VERIFY){
        fpData=(struct fp_print_data*)iData;
    }
}

bool FingerprintDevice::canIdentify(){
    bool rc=false;
    if(isFp){
        if(fpDevice==NULL){  //Uups!
            syslog(LOG_ERR,"FIXME: fpDevice NULL.");
            return rc;
        }            
        struct fp_dev *dev=fp_dev_open(fpDevice);
        if(!dev){
            syslog(LOG_ERR,"Could not open fpDevice.");
            return rc;
        }
        else{
            if(fp_dev_supports_identification (dev)!=0){
                rc=true;
            }
            fp_dev_close(dev);
        }
        return rc;
    }
    // libbsapi type devices support it always
    return true;
}

struct fp_dscv_dev *FingerprintDevice::getFpDevice(){
    return fpDevice;
}

bool FingerprintDevice::isFpDevice(){
    return isFp;
}

int FingerprintDevice::getData(void **d){
    int size=0;
    if(isFp){
        if((size=fp_print_data_get_data(fpData,(unsigned char**)d))==0){
            syslog(LOG_ERR,"Could not convert fpData!");
        }
    }
#ifdef HAS_LIBBSAPI
    else{
        *d=(void*)bsData;
        size=bsData->Header.Length;
    }
#endif    
    return size;
}

void FingerprintDevice::setData(void *d,int size){
    if(isFp)
        fpData=fp_print_data_from_data((unsigned char*)d,size);
#ifdef HAS_LIBBSAPI
    else
        bsData=(ABS_BIR*)d;
#endif    
}

// Here we go...
void FingerprintDevice::run(){
    syslog(LOG_DEBUG,"Starting %s.",(mode==MODE_VERIFY?"verify":mode==MODE_ACQUIRE?"acquire":"identify"));
    switch(mode){
        case MODE_VERIFY:
            verify();
            break;
        case MODE_ACQUIRE:
            acquire();
            break;
        case MODE_IDENTIFY:
            identify();
            break;
    }
    syslog(LOG_DEBUG,"Thread ended normally.");
}

void FingerprintDevice::stop(){
    if(isFp){
        if(dev!=NULL){
            // This is evil but needed
            // because of the missing "cancelOperation" call to libfprint devices:
            setTerminationEnabled(true);
            terminate();
            dev=NULL;
        }
        return;
    }
#ifdef HAS_LIBBSAPI
    else{
        if(conn==0){
            syslog(LOG_DEBUG,"No device to stop.");
            return;
        }
        ABSCancelOperation(conn,mode);
    }
#endif
}

// private helpers -------------------------------------------------------------
#ifdef HAS_LIBBSAPI
// These are helpers for the callback
// I need them only because I don't know how to handover a member function of
// FingerprintDevice as callback to the ABS... functions.
void FingerprintDevice::emitAcquireResult(int result){
    emit acquireResult(result);
}

void FingerprintDevice::emitVerifyResult(int result){
    emit verifyResult(result);
}

static FingerprintDevice *fingerprintDevice;
#endif

//returns an open fp_dev
struct fp_dev *fpDevOpen(struct fp_dscv_dev *fpDevice){
    struct fp_dev *device;

    if(fpDevice==NULL){
        syslog(LOG_ERR,"FIXME: fpDevice NULL.");
        return NULL;
    }
    device=fp_dev_open(fpDevice);
    if(!device){
        syslog(LOG_ERR,"Could not open fpDevice.");
        return NULL;
    }
    return device;
}

#ifdef HAS_LIBBSAPI
//opens an ABS... device
bool bsDevOpen(string bsDevice,ABS_CONNECTION *conn){
    if(bsDevice.length()==0){
        syslog(LOG_ERR,"FIXME: BsDevice string empty.");
        return false;
    }
        
    ABS_STATUS res=ABSOpen(bsDevice.data(),conn);
    if(res!=ABS_STATUS_OK) {
        const ABS_CHAR *ErrorMessage;
        ABS_DWORD ErrorCode;
        ABSGetLastErrorInfo(&ErrorCode,&ErrorMessage);
        syslog(LOG_ERR,"ABSOpen() failed %d (%s).",(int)ErrorCode,ErrorMessage);
        return false;
    }
    return true;
}

void retry(const ABS_OPERATION* p_operation){
    switch(p_operation->OperationID){
        case MODE_ACQUIRE:
            fingerprintDevice->emitAcquireResult(FP_ENROLL_RETRY);
            break;
        case MODE_VERIFY:
        case MODE_IDENTIFY:
            fingerprintDevice->emitVerifyResult(FP_VERIFY_RETRY);
    }
}

void center(const ABS_OPERATION* p_operation){
    switch(p_operation->OperationID){
        case MODE_ACQUIRE:
            fingerprintDevice->emitAcquireResult(FP_ENROLL_RETRY_CENTER_FINGER);
            break;
        case MODE_VERIFY:
        case MODE_IDENTIFY:
            fingerprintDevice->emitVerifyResult(FP_VERIFY_RETRY_CENTER_FINGER);
    }
}

void tooShort(const ABS_OPERATION* p_operation){
    switch(p_operation->OperationID){
        case MODE_ACQUIRE:
            fingerprintDevice->emitAcquireResult(FP_ENROLL_RETRY_TOO_SHORT);
            break;
        case MODE_VERIFY:
        case MODE_IDENTIFY:
            fingerprintDevice->emitVerifyResult(FP_VERIFY_RETRY_TOO_SHORT);
    }
}

static ABS_DWORD lastMsg=-1;

void BSAPI callback(const ABS_OPERATION* p_operation,ABS_DWORD msg,void* /*data*/){
    switch(msg) {
        case ABS_MSG_PROCESS_SUCCESS:
            if(lastMsg==ABS_MSG_PROMPT_SCAN&&p_operation->OperationID==MODE_ACQUIRE){
                fingerprintDevice->emitAcquireResult(FP_ENROLL_PASS);
                lastMsg=-1;
            }
            break;
        case ABS_MSG_PROCESS_FAILURE:
            syslog(LOG_DEBUG,"ABS_MSG_PROCESS_FAILURE.");
            break;
        /* Prompt messages should inform the user that he should do something. */
        case ABS_MSG_PROMPT_SCAN:
            lastMsg=msg;
            syslog(LOG_DEBUG,"swipe the finger."); 
            break;
        case ABS_MSG_PROMPT_TOUCH:
            syslog(LOG_DEBUG,"touch the sensor.");
            break;
        case ABS_MSG_PROMPT_KEEP:
            syslog(LOG_DEBUG,"keep finger on the sensor."); 
            break;
        case ABS_MSG_PROMPT_LIFT:
            syslog(LOG_DEBUG,"lift your finger away from the sensor.");
            break;
        case ABS_MSG_PROMPT_CLEAN:
            syslog(LOG_DEBUG,"clean the sensor."); 
            break;
        
        /* Quality messages come if something went wrong. E.g. the user
         * did not scan his finger in the right way. */
        case ABS_MSG_QUALITY_CENTER_HARDER:
            syslog(LOG_DEBUG,"bad quality: center and harder."); 
            center(p_operation);
            break;
        case ABS_MSG_QUALITY_CENTER:
            syslog(LOG_DEBUG,"bad quality: center."); 
            center(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_LEFT:
            syslog(LOG_DEBUG,"bad quality: too left."); 
            center(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_RIGHT:
            syslog(LOG_DEBUG,"bad quality: too right."); 
            center(p_operation);
            break;
        case ABS_MSG_QUALITY_HARDER:
            syslog(LOG_DEBUG,"bad quality: harder."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_LIGHT:
            syslog(LOG_DEBUG,"bad quality: too light."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_DRY:
            syslog(LOG_DEBUG,"bad quality: too dry.");
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_SMALL:
            syslog(LOG_DEBUG,"bad quality: too small.");
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_SHORT:
            syslog(LOG_DEBUG,"bad quality: too short."); 
            tooShort(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_HIGH:
            syslog(LOG_DEBUG,"bad quality: too high."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_LOW:
            syslog(LOG_DEBUG,"bad quality: too low."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_FAST:
            syslog(LOG_DEBUG,"bad quality: too fast."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_SKEWED:
            syslog(LOG_DEBUG,"bad quality: too skewed."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_TOO_DARK:
            syslog(LOG_DEBUG,"bad quality: too dark."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_BACKWARD:
            syslog(LOG_DEBUG,"bad quality: backward movement detected."); 
            retry(p_operation);
            break;
        case ABS_MSG_QUALITY_JOINT:
            syslog(LOG_DEBUG,"bad quality: joint detected."); 
            retry(p_operation);
            break;
            
//UNUSED MESSAGES:
//        /* These messages just inform us how the interactive operation
//         * progresses. Typical applications do not need it. */
//        case ABS_MSG_PROCESS_BEGIN:
//        case ABS_MSG_PROCESS_END:
//        /* On some platforms, the biometric operastion can be suspended
//         * when other process acquires sensor for other operation. */
//        case ABS_MSG_PROCESS_SUSPEND:
//        case ABS_MSG_PROCESS_RESUME:
//        /* Sometimes some info how the operation progresses is sent. */
//        case ABS_MSG_PROCESS_PROGRESS:
//        /* Navigation messages are sent only from ABSNavigate. Its not used
//         * in this sample but we list the messages here for completeness. */
//        case ABS_MSG_NAVIGATE_CHANGE:
//        case ABS_MSG_NAVIGATE_CLICK:
//        /* Real application would probably use some GUI to provide feedback
//         * for user. On these messages the GUI dialog should be made vsiible
//         * and invisible respectivelly. */
//        case ABS_MSG_DLG_SHOW:
//        case ABS_MSG_DLG_HIDE:
    }
}

static ABS_OPERATION operation={ 
    0,          //operation identifier, used for mode (acquire, identify or verify)
    NULL,       //data to call back, not used
    callback,  
    OPERATION_TIMEOUT,      //timeout
    ABS_OPERATION_FLAG_LL_CALLBACK
};
#endif    

void FingerprintDevice::setTimeout(bool
#ifdef HAS_LIBBSAPI
    timeout){
    if(timeout){
        operation.Timeout=OPERATION_TIMEOUT;
    }
    else{
        operation.Timeout=0;
    }
#else
    ){
#endif
}

//run a verification task
bool FingerprintDevice::verify(){
    struct fp_img *img=NULL;
    int result;
    struct fp_dev *dev=NULL;
#ifdef HAS_LIBBSAPI
    ABS_LONG bsResult;
//    ABS_SAMPLE_IMAGE *bsImage;
#endif    

    if(isFp&&fpDevice!=NULL){
        dev=fpDevOpen(fpDevice);
    }
#ifdef HAS_LIBBSAPI
    else{
        if(bsDevice.length()!=0)
            if(!bsDevOpen(bsDevice,&conn))
                conn=0;
        fingerprintDevice=this;
    }
    if(dev==NULL&&conn==0){
#endif    
#ifndef HAS_LIBBSAPI
    if(dev==NULL){
#endif    
        syslog(LOG_ERR,"No device open."); 
        if(fpData!=NULL){
            fp_print_data_free(fpData);
            fpData=NULL;
        }
        emit noDeviceOpen();
        return false;
    }
    emit neededStages(1);   //always "1" for verification
    if(isFp){
        do{
            sleep(1);   //don't know what this is good for; found it in some sample code of libfprint
            result=fp_verify_finger_img(dev,fpData,&img);
            //"img" to be used in later version
//            if(img){
//                syslog(LOG_DEBUG,"Verify: have image.");
//                fp_img_free(img);
//            }
            if(result<0){
                syslog(LOG_ERR,"Verify failed with error %d.",result); 
            }
            emit verifyResult(result);
        }while(result!=FP_VERIFY_NO_MATCH&&result!=FP_VERIFY_MATCH);
        //This is for identifying with "no idetifying" devices
        if(result==FP_VERIFY_MATCH)
            emit matchResult(0);
        else
            emit matchResult(-1);
        
        fp_dev_close(dev);
        dev=NULL;
        fp_print_data_free(fpData);
        fpData=NULL;
        return true;
    }
#ifdef HAS_LIBBSAPI
    else{           //is bsDevice
        operation.OperationID=MODE_VERIFY;
        if(ABSVerify(conn,&operation,1,&bsData,&bsResult,0)!=ABS_STATUS_OK){
            syslog(LOG_ERR,"ABSVerify() failed."); 
            emit verifyResult(FP_VERIFY_NO_MATCH);
            return false;
        }
        //insert "image" stuff here in later version
//        syslog(LOG_DEBUG,"Verify: grab image start.");
//        if(ABSGrab(conn,&operation,ABS_PURPOSE_VERIFY,&bsImage,0)==ABS_STATUS_OK){
//            syslog(LOG_DEBUG,"Verify: have image.");
//            ABSFree(bsImage);
//        }
//        syslog(LOG_DEBUG,"Verify: grab image end.");



        if(ABSClose(conn)!=ABS_STATUS_OK) {
            syslog(LOG_ERR,"ABSClose() failed."); 
        }
        if(bsResult==0){
            emit verifyResult(FP_VERIFY_MATCH);
            return true;
        }
        emit verifyResult(FP_VERIFY_NO_MATCH);
    }
#endif
    return false;
}

//run an enrollment task
bool FingerprintDevice::acquire(){
    int result;
    //libfprint stuff
    struct fp_img *img=NULL;
    mode=MODE_ACQUIRE;
    if(fpDevice!=NULL)
        dev=fpDevOpen(fpDevice);
#ifdef HAS_LIBBSAPI
    //libbsapi stuff
    if(bsDevice.length()!=0)
        if(!bsDevOpen(bsDevice,&conn))
            conn=0;
    fingerprintDevice=this;    
    if(dev==NULL&&conn==0){
#endif    
#ifndef HAS_LIBBSAPI
    if(dev==NULL){
#endif    
        syslog(LOG_ERR,"No device open."); 
        emit noDeviceOpen();
        return false;
    }

    if(isFp)
        emit neededStages(fp_dev_get_nr_enroll_stages(dev));
    else                //is bsDevice
        emit neededStages(3);       //don't know how to get this from libbsapi

    if(isFp){
        do{
            sleep(1);   //don't know what this is good for; found it in some sample code of libfprint
            result=fp_enroll_finger_img(dev,&fpData,&img);
            //"img" to be used in later version
            if(img){
                syslog(LOG_DEBUG,"Acquire: have image.");
                fp_img_free(img);
            }
            if(result<0){
                syslog(LOG_ERR,"Acquire failed with error %d.",result); 
            }
            else{
                syslog(LOG_DEBUG,"Acquire result %d.",result); 
            }
            emit acquireResult(result);
        }while(result!=FP_ENROLL_COMPLETE);
        fp_dev_close(dev);
        dev=NULL;
    }
#ifdef HAS_LIBBSAPI
    else{           //is bsDevice
        operation.OperationID=mode;
        if(ABSEnroll(conn,&operation,&bsData,0)!=ABS_STATUS_OK){
            syslog(LOG_ERR,"ABSEnroll() failed."); 
            emit acquireResult(FP_ENROLL_FAIL);
            return false;
        }
        //insert "image" stuff here in later version
        if(ABSClose(conn)!=ABS_STATUS_OK) {
            syslog(LOG_ERR,"ABSClose() failed."); 
        }
        emit acquireResult(FP_ENROLL_COMPLETE);
        return true;
    }
#endif
    return true;
}

bool FingerprintDevice::identify(){
    long match=-1L;

    if(identifyData==NULL||numIdentify==0){
        syslog(LOG_ERR,"No data to identify."); 
        return false;
    }
    if(isFp){
        if(fpDevice!=NULL)
            dev=fpDevOpen(fpDevice);
        if(dev==NULL){
            syslog(LOG_ERR,"Could not open libfprint-device."); 
            emit noDeviceOpen();
            return false;
        }
        int result=fp_identify_finger(dev,(struct fp_print_data **)identifyData,(size_t*)&match);
        if(result<0){
            syslog(LOG_ERR,"Identify failed with error %d.",result); 
            match=-1L;
        }
        fp_dev_close(dev);
        dev=NULL;
    }
#ifdef HAS_LIBBSAPI
    else{           //is bsDevice
        if(bsDevice.length()!=0)
            if(!bsDevOpen(bsDevice,&conn))
                conn=0;
        if(conn==0){
            syslog(LOG_ERR,"Could not open libbsapi-device."); 
            emit noDeviceOpen();
            return false;
        }
        fingerprintDevice=this;
        operation.OperationID=MODE_IDENTIFY;
        if(ABSVerify(conn,&operation,numIdentify,(ABS_BIR**)identifyData,(ABS_LONG*)&match,absFlag)!=ABS_STATUS_OK){
            const ABS_CHAR *ppErrorMessage;
            ABS_DWORD pErrorCode;
            ABSGetLastErrorInfo(&pErrorCode,&ppErrorMessage);
            syslog(LOG_ERR,"ABSVerify() failed with %s.",ppErrorMessage); 
        }
        if(ABSClose(conn)!=ABS_STATUS_OK) {
            syslog(LOG_ERR,"ABSClose() failed."); 
        }
    }
#endif
    emit matchResult(match);
    return true;
}
