/*
 * Project "fingerprintGUI": Services for fingerprint authentication on Linux
 * Module: FingerprintDiscoverer.cpp, FingerprintDiscoverer.h
 * Purpose: Try to load all stored fingerprints of all users on local machine
 *          Doesn't load fingerprint data from homedirs where it has no permission to read
 *          So only running it as root can indentify all users.
 *
 * @author  Wolfgang Ullrich
 * Copyright (C) 2008-2009  Wolfgang Ullrich
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sstream>
#include <iostream>
#include <fstream>
#include <string>
#include <QtGui>
#include <qstring.h>
#include <qwidget.h>
#include <string>
#include <sys/stat.h>

#include "FingerprintDevice.h"
#include "fingercodes.h"
#include "DeviceHandler.h"
#include "FingerprintDiscoverer.h"

typedef enum {
    PWD_USER    = 0,
    PWD_PASS    = 1,
    PWD_UID     = 2,
    PWD_GID     = 3,
    PWD_NAME    = 4,
    PWD_HOME    = 5,
    PWD_LOGIN   = 6
} pwdTokens;

//#define PWD_TOKEN ":"
#define PWD_TOKEN ':'

string strToLower(string strToConvert){
   for(unsigned int i=0;i<strToConvert.length();i++){
      strToConvert[i] = tolower(strToConvert[i]);
   }
   return strToConvert;
}

//constructor to be used for fingerprintIdentifier (for all users)
FingerprintDiscoverer::FingerprintDiscoverer(FingerprintDevice *d,char *passwdPath){
    string line;
    char *pwdToken[7];
    struct stat st;
    device=d;
    numPrints=0;
    allFingerprints=NULL;

    syslog(LOG_INFO,"Collecting users/homedirs from %s.",passwdPath);
    ifstream pwdFile (passwdPath);
    if(pwdFile.is_open()){
        while(!pwdFile.eof()){
            getline(pwdFile,line);
            if(line.length()!=0&&line[0]!='#'){  //skip comment lines and empty lines
//syslog(LOG_DEBUG,"Line: %s",line.data());
                char* tok=(char*)line.data();
                for(int i=0;i<7;i++){
                    pwdToken[i]=tok;
                    char* end=strchr(tok,PWD_TOKEN);
                    if(end==NULL) break;
                    *end='\0';
                    tok=++end;
                }
                if(pwdToken[PWD_USER]!=NULL&&pwdToken[PWD_HOME]!=NULL&&pwdToken[PWD_LOGIN]!=NULL){
                    string token=pwdToken[PWD_LOGIN];
                    token=strToLower(token);
                    if(token.find("false")==std::string::npos&&
                       token.find("none")==std::string::npos&&
                       token.find("null")==std::string::npos){  //user has login
                        if(stat(pwdToken[PWD_HOME],&st)==0){    //home exists
                            syslog(LOG_DEBUG,"Have user: %s with home: %s",pwdToken[PWD_USER],pwdToken[PWD_HOME]);
                            discoverFingerprints(pwdToken[PWD_USER],pwdToken[PWD_HOME],device);
                        }
                    }
                }
            }
        }
        pwdFile.close();
    }
    else{
        syslog(LOG_ERR,"Could not open %s.",passwdPath);
        return;
    }
   setupArrays();
}

//constructor to be used for fingerprintPAM (for a known user only)
FingerprintDiscoverer::FingerprintDiscoverer(FingerprintDevice *d, string user, string home){
    device=d;
    numPrints=0;
    allFingerprints=NULL;

    discoverFingerprints(user,home,d);
    setupArrays();
}

// public getters and setters --------------------------------------------------
// returns array of fingerprintData objects to get the finger names
FingerprintData **FingerprintDiscoverer::getFingerprintData(int *numData){
    *numData=numPrints;
    return  fingerprintData;
}

// returns array of bir data to be identified
void **FingerprintDiscoverer::getIdentifyData(int *numData){
    *numData=numPrints;
    return  identifyData;
}

// private helpers -------------------------------------------------------------
void FingerprintDiscoverer::setupArrays(){
    if(numPrints==0){
        syslog(LOG_ERR,"Found no fingerprints for this device! Aborting.");
        return;
    }
    syslog(LOG_DEBUG,"Found %d fingerprints for this device.",numPrints);
    identifyData=new void*[numPrints+1];
    fingerprintData=new FingerprintData*[numPrints];
    FingerprintData *fpd=allFingerprints;
    for(int i=0;i<numPrints;){
        if(fpd->isFpDevice()){
            identifyData[i]=fpd->getFpData();
        }
#ifdef HAS_LIBBSAPI
        else{
            identifyData[i]=fpd->getBsData();
        }
#endif
        fingerprintData[i]=fpd;
        i++;
        identifyData[i]=NULL;
        fpd=fpd->next;
    }
    syslog(LOG_DEBUG,"Arrays are setup now.");
}

void FingerprintDiscoverer::addFingerprintData(FingerprintData *fpd){
    syslog(LOG_DEBUG,"Adding %s -- %s(%s).",fpd->getUserName()->data(),fpd->getFingerName(),(fpd->isFpDevice()?"(FP)":"(BS)"));
    if(allFingerprints==NULL){
        allFingerprints=fpd;
        fpd->next=NULL;
        return;
    }
    for(FingerprintData *fp=allFingerprints;fp!=NULL;fp=fp->next){
        if(fp->next==NULL){
            fp->next=fpd;
            fpd->next=NULL;
            return;
        }
    }
}

// discover stored fingerprints fitting this device for an user in his homedir
void FingerprintDiscoverer::discoverFingerprints(string user,string homedir,FingerprintDevice *d){
    string fname=string(homedir);
    fname.append(DATA_DIR);
    if(d->isFpDevice()){
        fname.append(FP_DATA_DIR);
    }
#ifdef HAS_LIBBSAPI
    else{
        fname.append(BS_DATA_DIR);
    }
#endif
    std::ifstream check(fname.data());
    if(!check){                         //directory does not exists
        syslog(LOG_DEBUG,"Could not open %s",fname.data());
        return;
    }
    for(int f=0;f<10;f++){
        std::ostringstream o;
        o<<f;
        string fileName=string(fname);
        fileName.append(o.str());
        fileName.append(DATA_EXT);
        ifstream birFile(fileName.data(),ios_base::binary);
        if(birFile.is_open()){
            birFile.seekg(0,ios::end);
            int dataSize=birFile.tellg();
            if(dataSize==0){
                birFile.close();
                continue;
            }
            birFile.seekg(0,ios::beg);
            syslog(LOG_DEBUG,"Reading %s (%d byte).",fileName.data(),dataSize);
            char *birData=new char[dataSize];
            birFile.read((char*)birData,dataSize);
            birFile.close();
            FingerprintData *fpd=NULL;
            if(d->isFpDevice()){
                fp_print_data *fp=fp_print_data_from_data((unsigned char*)birData,dataSize);
                if(fp==NULL){
                    syslog(LOG_ERR,"Could not convert fpData!");
                    return;
                }
                fpd=new FingerprintData(user,f,fp);
            }
#ifdef HAS_LIBBSAPI
            else{
                fpd=new FingerprintData(user,f,(ABS_BIR*)birData);
            }
#endif
            if(fpd)
                addFingerprintData(fpd);
            numPrints++;
        }
    }
}
