/*
 * Project "fingerprintGUI": Services for fingerprint authentication on Linux
 * Module: PamGUI.cpp, PamGUI.h
 * Purpose: Main object for fingerprintPAM module for running in gui environments
 *
 * @author  Wolfgang Ullrich
 * Copyright (C) 2008-2009  Wolfgang Ullrich
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "PamGUI.h"

PamGUI::PamGUI(FingerprintDevice *dev,const char *user,const char *finger,QWidget *parent,Qt::WFlags f)
    : QFrame(parent, f){
    setupUi(this);

    fifo=-1;
    if(user!=NULL){
        QString s="Authenticating ";
        s.append(user);
        showMessage(false,s);
    }
    if(finger!=NULL){
        string s=string("Please swipe your ");
        s.append(finger);
        s.append(".");
        label->setText(QApplication::translate("MainWindow",s.data(),0,QApplication::UnicodeUTF8));
        syslog(LOG_DEBUG,"Label: %s",s.data());
    }
    fingerprintData=NULL;
    startupGUI(dev);
}

PamGUI::PamGUI(FingerprintDevice *dev,void **iData,FingerprintData **fData,QWidget *parent,Qt::WFlags f)
    : QFrame(parent, f){
    setupUi(this);

    fifo=-1;
    showMessage(false,"Ready...");
    identifyData=iData;
    fingerprintData=fData;
    startupGUI(dev);
}

void PamGUI::startupGUI(FingerprintDevice *dev){
    UNDECORATED_MODAL_WINDOW_CENTERED
    ANIMATION_MOVIE    
    show();
    repeatDelay=0;
    device=dev;
    device->setTimeout(false);
    connect(device,SIGNAL(verifyResult(int)),this,SLOT(newVerifyResult(int)));
    connect(device,SIGNAL(matchResult(int)),this,SLOT(matchResult(int)));
    device->start();
    releaseKeyboard();
    START_TIMER
}

void PamGUI::showMessage(bool bold,const QString msg){
    QFont font;
    if(bold){
        font.setPointSize(10);
        font.setBold(true);
        statusBar->setFont(font);
    }
    else{
        font.setPointSize(9);
        font.setBold(false);
        statusBar->setFont(font);
    }
    statusBar->setText(msg);    // Show meqt named pipe examplessage on statusBar
    syslog(LOG_DEBUG,"Message: %s",string(msg.toLatin1()).data()); // Message to syslog
    
    // send message to plugin
    if(fifo<0){     // Try to open named pipe
        fifo=open(PLUGIN_FIFO,O_WRONLY|O_NDELAY);
        if(fifo<0){ // Maybe there is no fingerprintPlugin listening
            syslog(LOG_DEBUG,"Could not open named pipe (fingerprintPlugin not present).");
            return;
        }
        else syslog(LOG_DEBUG,"Opened fifo to fingerprintPlugin: %d",fifo);
    }  
    
    string fifoMsg(""); // Compose a message for fingerprintPlugin
    if(bold) fifoMsg.append(MSG_BOLD);
    else fifoMsg.append(MSG_NORMAL);
    fifoMsg.append(msg.toLatin1());
    fifoMsg.append("\0");
    if(write(fifo,fifoMsg.data(),strlen(fifoMsg.data())+1)<0)
        syslog(LOG_ERR,"Error writing to fifo: %s",strerror(errno));
    else
        syslog(LOG_DEBUG,"wrote to fifo: %s",fifoMsg.data());
}

// slots -----------------------------------------------------------------------
void PamGUI::matchResult(int match){
    device->stop();
    if(match>=0){
        movie->setPaused(true);
        string message;
        if(fingerprintData!=NULL){
            message.append("Identified: ");
            message.append(fingerprintData[match]->getUserName()->data());
            message.append(" (");
            message.append(fingerprintData[match]->getFingerName());
            message.append(")");
        }
        else{
            message.append("OK");
        }
        showMessage(true,message.data());
	timer->stop();
        //exit with index (match) as exit code
        qApp->processEvents();
        device->wait(5000);
        for(int i=SHOW_DELAY/1000;i>0;i--){ //let 'em see the result before exiting
            qApp->processEvents();
            usleep(1000);   
        }
        unlink(PLUGIN_FIFO);  // We don't need the fifo any more
        qApp->exit(match);
        return;
    }
    showMessage(true,"Not identified!");
    repeatDelay=2;   //let 'em see the result before restarting
}

void PamGUI::newVerifyResult(int result){
    switch(result){
        case FP_VERIFY_NO_MATCH:
            showMessage(false,"No match!");
            break;
        case FP_VERIFY_RETRY_TOO_SHORT:
            showMessage(false,"Swipe too short...");
            break;
        case FP_VERIFY_RETRY_CENTER_FINGER:
            showMessage(false,"Please center...");
            break;
        case FP_VERIFY_RETRY:
        case FP_VERIFY_RETRY_REMOVE_FINGER:
            showMessage(false,"Try again...");
            break;
        default:
            return;
    }
}

// Helper thread for restart
void PamGUI::timerTick(){
    switch(repeatDelay){
        case 0:
            if(!device->isRunning()){
                syslog(LOG_ERR,"ERROR: Fingerprint device not running.");
                qApp->exit(-1);
                return;
            }
            //do nothing
            break;
        case 1:
            syslog(LOG_INFO,"Waiting for device to stop...");
            device->wait(5000);
            syslog(LOG_INFO,"Stopped, restarting");
            //restart fingerprint scanner
            showMessage(false,"Ready...");
            device->start();
            repeatDelay--;
            break;
        default:
            repeatDelay--;  //still wait
    }
}
