/*
 * Project "fingerprintGUI": Services for fingerprint authentication on Linux
 * Module: Helper.cpp
 * Purpose: Helper process to be used for PAM authentication
 *
 * @author  Wolfgang Ullrich
 * Copyright (C) 2008-2009  Wolfgang Ullrich
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <QtGui>
#include <QObject>
#include <QEventLoop>

#include <sys/wait.h>
#include <linux/uinput.h>
#include <pwd.h>
#include <fakekey/fakekey.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <X11/keysym.h>

#include "../include/FingerprintDevice.h"
#include "../include/DeviceHandler.h"
#include "../include/FingerprintDiscoverer.h"
#include "../include/PamGUI.h"
#include "../include/PamNonGUI.h"
#include "../include/fingercodes.h"

#include "../include/globals.h"

using namespace std;

static string syslogIdent=string("fingerprintHelper");
static volatile bool userSent=false;

// Send char to uinput (only digits and enter are supported and needed here because we use it for "randomString" only)
bool uinputSendKey(char key,int uinput){
    struct input_event event;
    unsigned int keycode;

    switch(key){
        case '0':
            keycode=KEY_0;
            break;
        case '1':
            keycode=KEY_1;
            break;
        case '2':
            keycode=KEY_2;
            break;
        case '3':
            keycode=KEY_3;
            break;
        case '4':
            keycode=KEY_4;
            break;
        case '5':
            keycode=KEY_5;
            break;
        case '6':
            keycode=KEY_6;
            break;
        case '7':
            keycode=KEY_7;
            break;
        case '8':
            keycode=KEY_8;
            break;
        case '9':
            keycode=KEY_9;
            break;
        default:
            keycode=KEY_ENTER;
    }
    
    memset(&event,0,sizeof(event));
    // PRESS
    gettimeofday(&event.time,NULL);
    event.type=EV_KEY;
    event.code=keycode;
    event.value=1;
    if(write(uinput, &event, sizeof(event))<0){
        syslog(LOG_ERR,"Write to uinput failed (%s).",strerror(errno));
        return false;
    }
    event.type=EV_SYN;
    event.code=SYN_REPORT;
    event.value=0;
    if(write(uinput,&event,sizeof(event))<0){
        syslog(LOG_ERR,"Write to uinput failed (%s).",strerror(errno));
        return false;
    }
    // RELEASE
    gettimeofday(&event.time,NULL);
    event.type=EV_KEY;
    event.code=keycode;
    event.value=0;
    if(write(uinput, &event, sizeof(event))<0){
        syslog(LOG_ERR,"Write to uinput failed (%s).",strerror(errno));
        return false;
    }
    event.type=EV_SYN;
    event.code=SYN_REPORT;
    event.value=0;
    if(write(uinput,&event,sizeof(event))<0){
        syslog(LOG_ERR,"Write to uinput failed (%s).",strerror(errno));
        return false;
    }
    return true;
}

bool uinputWrite(const char *line,int uinput){
    for(unsigned int i=0;i<strlen(line);i++){
        if(!uinputSendKey(line[i],uinput))
            return false;
    }
    if(!uinputSendKey('\n',uinput))
        return false;
    return true;
}

// Sends the line and <enter> key to PAM prompt so PAM prompt exits with the given line
bool exitPrompt(const char *line,const char *display){
    struct uinput_user_dev dev;
    int uinput=-1;
    Display *xdpy;
    FakeKey *fakekey;
    
    if(display!=NULL){  // We have a X display, using libfakekey
        syslog(LOG_DEBUG,"Using libfakekey to exit PAM conversation (%s).",line);
        if((xdpy=XOpenDisplay(display))==NULL){
            syslog(LOG_ERR,"ERROR: XOpenDisplay.");
            return false;
        }
        if((fakekey=fakekey_init(xdpy))==NULL){
            syslog(LOG_ERR,"ERROR: Initializing FAKEKEY.");
            return false;
        }
        if(line!=NULL){
            int l=strlen(line);
            for(int i=0;i<l;i++){
                fakekey_press(fakekey,(const unsigned char*)&line[i],1,0);
                usleep(6000);   // Wait for X to process the key
                fakekey_release(fakekey);
                usleep(6000);   // Wait for X to process the key
            }
        }
        fakekey_press_keysym(fakekey,XK_Return,0);
        fakekey_release(fakekey);
    }
    else{   // No X display, using udev
        syslog(LOG_DEBUG,"Using uinput to exit PAM conversation.");
        uinput=open("/dev/input/uinput",O_WRONLY|O_NDELAY);
        if(uinput<0)
            uinput=open("/dev/misc/uinput",O_WRONLY|O_NDELAY);
        if(uinput<0)
            uinput=open("/dev/uinput",O_WRONLY|O_NDELAY);
        if(uinput<0){
            syslog(LOG_ERR,"ERROR: Open uinput failed (Do you have write permission?).");
            return false;
        }
        memset(&dev,0,sizeof(dev));
        strncpy(dev.name,"fingerprintPAM",UINPUT_MAX_NAME_SIZE);
        dev.id.version=4;
        dev.id.bustype=BUS_USB;
        ioctl(uinput,UI_SET_EVBIT,EV_KEY);
        ioctl(uinput,UI_SET_EVBIT,EV_REP);
        ioctl(uinput,UI_SET_KEYBIT,KEY_ENTER);
        for(int i=0;i<10;i++)
            ioctl(uinput,UI_SET_KEYBIT,KEY_1+i);
        if(write(uinput,&dev,sizeof(dev))<0){
            syslog(LOG_ERR,"Write to uinput failed (%s).",strerror(errno));
            return false;
        }
        if(ioctl(uinput,UI_DEV_CREATE)){
            syslog(LOG_ERR,"Ioctl uinput failed (%s).",strerror(errno));
            return false;
        }
        return uinputWrite(line,uinput);
    }
    return true;
}

void stopPlugin(){
    int pluginFifo=open(PLUGIN_FIFO,O_WRONLY|O_NDELAY);   // In case we have a fingerprintPlugin running, stop it
    if(pluginFifo<0){ // Maybe there is no fingerprintPlugin listening
        syslog(LOG_DEBUG,"Could not open named pipe (fingerprintPlugin not present).");
    }
    else{
        syslog(LOG_DEBUG,"Opened fifo to fingerprintPlugin: %d",pluginFifo);
        string fifoMsg(MSG_STOP); // Compose a message for fingerprintPlugin
        if(write(pluginFifo,fifoMsg.data(),strlen(fifoMsg.data())+1)<0)
            syslog(LOG_ERR,"ERROR: Writing to fifo: %s",strerror(errno));
        else
            syslog(LOG_DEBUG,"Wrote to fifo: %s",fifoMsg.data());
    }
}

// Catch SIGUSR1. When SIGUSR1 is raised and we have _not_ sent a username to the parent,
// we can exit because the user has typed the password by keyboard.
static void handler_SIGUSR1(int sig){
    switch(sig){
        case SIGUSR1:
            syslog(LOG_DEBUG,"Got SIGUSR1, %s.",userSent?"continueing":"EXIT_SUCCESS");
            if(!userSent){  // We did not send a username to parent
                stopPlugin();
                closelog();
                exit(EXIT_SUCCESS);
            }
            return;
        case SIGUSR2:
            syslog(LOG_DEBUG,"Got SIGUSR2, EXIT_SUCCESS.");
            exit(EXIT_SUCCESS);
    }
    syslog(LOG_DEBUG,"Got signal %d, ignoring.",sig);
}

// Request fingerprint while PAM propmts for username/password
// Returns false on error, true on success
bool requestFingerprint(const char *display,char *username,char *passwdPath,int argc,char **argv){
    int lastResult,numIdentify;
    void **identifyData=NULL;
    FingerprintData **fingerprintData=NULL;
    const char *xauthpath=NULL;

    //init libfprint
    if(fp_init()!=0){
        syslog(LOG_ERR,"Unable to init libfprint.");
        return false;
    }
    syslog(LOG_DEBUG,"Libfprint initialized.");

    //init libbsapi
#ifdef HAS_LIBBSAPI
    syslog(LOG_INFO,"Have libbsapi.");
    if(ABSInitialize()!=0){
        syslog(LOG_ERR,"Unable to init libbsapi.");
        return false;
    }
    syslog(LOG_DEBUG,"Libbsapi initialized.");
#endif
  
    // Find fingerprint device to be used
    // If no identifier device is found we fallback to verifying the first fingerprint discovered for this user
    FingerprintDevice *devices=NULL;
    int mode=MODE_IDENTIFY;
    DeviceHandler deviceHandler(DISPLAY_DRIVER_NAME);
    deviceHandler.rescan();
    devices=deviceHandler.getIdentifiers();
    
    if(devices==NULL){
        mode=MODE_VERIFY;
        syslog(LOG_WARNING,"WARNING: No devices can identify. Fallback to verifying only.");
        devices=deviceHandler.getVerifiers();
        if(devices==NULL){
            syslog(LOG_WARNING,"No useful devices found.");
            return false;
        }
    }
    if(devices->next!=NULL){
        syslog(LOG_WARNING,"Found more then one devices. Using %s.",devices->getDisplayName(DISPLAY_DRIVER_NAME)->data());
    }

    xauthpath=getenv("XAUTHORITY");
    syslog(LOG_DEBUG,"Have display=%s, XAUTHORITY=%s.",display,xauthpath);
    
    if(username==NULL){             // user not known yet, maybe it's a login
        if(mode==MODE_IDENTIFY){    // device can identify. So identifying the user is possible
            //collect fingerprints from all users for this device
            FingerprintDiscoverer discoverer(devices,passwdPath);
            identifyData=discoverer.getIdentifyData(&numIdentify);
            fingerprintData=discoverer.getFingerprintData(&numIdentify);
            if(numIdentify<1){      // We have no fingerprints at all
                syslog(LOG_ERR,"No fingerprintData!");
                return false;
            }
            devices->setMode(MODE_IDENTIFY);
            devices->setIdentifyData(identifyData,numIdentify);

            if(display==NULL||xauthpath==NULL){
                //  NonGUI fingerprint identification is not possible! here
                //  If we would implement this, the first tty would open the fingerprint device
                //  and GUI login would never be possible.
                syslog(LOG_WARNING,"Missing %s (required for login).",display==NULL?"X-display":"XAUTHPATH");
                return false;
            }
            else{
                syslog(LOG_INFO,"Have X-display %s. Starting GUI login.",display);
                QApplication app(argc,argv);
                PamGUI gui(devices,identifyData,fingerprintData);
                if(!devices->isRunning()){  //something went wrong
                    syslog(LOG_ERR,"ERROR: Device not running!");
                    return false;
                }
                lastResult=app.exec();
            }
            if(lastResult>=0){
                const char *user=fingerprintData[lastResult]->getUserName()->data();
                syslog(LOG_DEBUG,"Have index %d (user: %s).",lastResult,user);
                userSent=true;
                exitPrompt(user,display);
                syslog(LOG_DEBUG,"Waiting for SIGUSR1.");
                sleep(1);
                syslog(LOG_DEBUG,"Waiting for PAM to be in password prompt.");
                usleep(100000);   //We need to wait for PAM being in the password prompt
                return true;
            }
            else{           // Something went wrong
                syslog(LOG_ERR,"ERROR: GUI exited for unknown reason!");
                return false;
            }
        }
        syslog(LOG_WARNING,"User unknown and device cannot identify!");
        return false;
    }
    
    // At this point:
    // The user was known already (authentication only)
    // Find user's homedir
    char homedir[FILENAME_MAX];
    struct passwd *passwd;
    passwd=getpwnam(username);
    if(!passwd){
        syslog(LOG_ERR,"No pwd entry for user %s.",username);
        return false;
    }
    sprintf(homedir,"%s",passwd->pw_dir);
    syslog(LOG_INFO,"Authenticating USER: %s, HOME: %s",username,homedir);

    // Collect fingerprints for this user/device in homedir
    FingerprintDiscoverer discoverer(devices,username,homedir);
    identifyData=discoverer.getIdentifyData(&numIdentify);
    if(mode==MODE_VERIFY){
        fingerprintData=discoverer.getFingerprintData(&numIdentify);
        if(fingerprintData==NULL){
            syslog(LOG_WARNING,"No fingerprintData for this user.");
            return false;
        }
    }
    if(numIdentify<1){
        syslog(LOG_ERR,"No fingerprintData for this user.");
        return false;
    }

    const char *fingername=NULL;
    devices->setMode(mode);
    if(mode==MODE_VERIFY){
        fingername=fingerprintData[0]->getFingerName(); // Use the first fingerprint found for this user
        devices->setIdentifyData((void**)fingerprintData[0]->getFpData(),0);
        syslog(LOG_DEBUG,"Verifying %s.",fingername);
    }
    else{
        devices->setIdentifyData(identifyData,numIdentify);
        syslog(LOG_DEBUG,"Identifying all discovered fingerprints from %s.",username);
    }

    if(display==NULL||xauthpath==NULL){  // We can't use the GUI
        syslog(LOG_WARNING,"WARNING: No %s. Starting command line authentication.",display==NULL?"X-display":"XAUTHPATH");
        QCoreApplication app(argc,argv);
        PamNonGUI nonGui(devices,username,fingername);
        if(!devices->isRunning()){  //something went wrong
            syslog(LOG_ERR,"ERROR: Device not running!");
            return false;
        }
        lastResult=app.exec();
    }
    else{
        syslog(LOG_INFO,"Have X-display %s. Starting GUI authentication.",display);
        QApplication app(argc,argv);
        PamGUI gui(devices,username,fingername);
        if(!devices->isRunning()){  //something went wrong
            syslog(LOG_ERR,"ERROR: Device not running!");
            return false;
        }
        lastResult=app.exec();
    }
    if(lastResult>=0){  // Fingerprint recognized
        return true;
    }
    return false;
}

// This is called via "execl" from fingerprintPAM
// arg[0] -- "fingerprintHelper"
// arg[1] -- pipe to fingerprintPAM (for sending the username)
// more args...
//  "-d", "--debug" for debug
//  "-u <name>", "--user <name>" for username (if known already by PAM)
//  "-p <path>", "--passwd <path>" for path to "passwd" file (default: /etc/passwd)

int main(int argc,char** argv) {
    int pipe;
    char *username=NULL;
    char *passwdPath=NULL;
    const char *display=getenv("DISPLAY");
    char randomString[11];

    openlog(syslogIdent.data(),LOG_NDELAY|LOG_PID,LOG_AUTH);
    setlogmask(LOG_UPTO(LOG_ERR));
    for(int i=0;i<argc;i++){
        if((strcmp(argv[i],"debug")==0)
                | (strcmp(argv[i],"--debug")==0)
                | (strcmp(argv[i],"-d")==0)){
            setlogmask(-1);
            syslog(LOG_INFO,"Got \"debug\" argument.");
        }
        if((strcmp(argv[i],"--user")==0)
                | (strcmp(argv[i],"-u")==0)){
            username=argv[++i];
            syslog(LOG_DEBUG,"Have username %s.",username);
        }
        if((strcmp(argv[i],"--passwd")==0)
                | (strcmp(argv[i],"-p")==0)){
            passwdPath=argv[++i];
            syslog(LOG_DEBUG,"Have passwd path %s.",passwdPath);
        }
    }
    syslog(LOG_ERR,"Started.");

    for(int i=0;i<argc;i++){
        syslog(LOG_DEBUG,"Arg%d: %s.",i,argv[i]);
    }

    if(argc<3){
        syslog(LOG_ERR,"Wrong number of arguments.\nIt's not intended to start fingerprintHelper manually.\n");
        return false;
    }
    
    signal(SIGUSR1,handler_SIGUSR1);    // Install a handler for SIGUSR1
    
    sscanf(argv[1],"%d",&pipe); // Get the pipe handle for IPC with fingerprintPAM
    syslog(LOG_DEBUG,"Have pipe %d.",pipe);
    memset(randomString,'\0',sizeof(randomString));
    if(read(pipe,randomString,sizeof(randomString))<=0){
        syslog(LOG_DEBUG,"ERROR: Read random string (%s).",strerror(errno));
        return false;
    }
    if(username==NULL)
        syslog(LOG_DEBUG,"Have no username.");
    if(passwdPath==NULL){
        syslog(LOG_DEBUG,"Have default passwd file.");
        passwdPath=PASSWD_FILE;
    }
        
    // If another helper process is running kill him
    FILE *pidfile=fopen(HELPER_PID,"r");
    if(pidfile){
        char pidstr[100];
        int i=fread(pidstr,1,100,pidfile);
        if(i>0)
            pidstr[i]='\0';
        sscanf(pidstr,"%d",&i);
        if(i!=getpid()){
            syslog(LOG_WARNING,"Found another helper process %d, killing it.",i);
            kill(i,SIGTERM);
        }
        fclose(pidfile);
    }
    pidfile=fopen(HELPER_PID,"w");
    if(pidfile){
        if(fprintf(pidfile,"%d",getpid())>0){
            syslog(LOG_DEBUG,"Wrote pidfile %s.",HELPER_PID);
            fclose(pidfile);
            chmod(HELPER_PID,0666);
        }
    }
    else
        syslog(LOG_WARNING,"Could not write to pidfile.");
    
    if(!requestFingerprint(display,username,passwdPath,argc,argv)){
        unlink(HELPER_PID);
#ifdef HAS_LIBBSAPI
        ABSTerminate();
#endif
        syslog(LOG_ERR,"ERROR: Requesting fingerprint, EXIT_FAILURE.");
	sleep(1);   //Maybe there is a plugin that is not ready for the stop command yet.
        stopPlugin();
        closelog();
        exit(EXIT_FAILURE);
    }
#ifdef HAS_LIBBSAPI
    ABSTerminate();
#endif
    exitPrompt(randomString,display);
    syslog(LOG_INFO,"Fingerprint recognized, EXIT_SUCCESS.");
    stopPlugin();
    closelog();
    unlink(HELPER_PID);
    exit(EXIT_SUCCESS);
}
