/*
 * Project "fingerprintGUI": Services for fingerprint authentication on Linux
 * Module: FingerprintPAM.cpp
 * Purpose: Main library to be used for PAM authentication
 *
 * @author  Wolfgang Ullrich
 * Copyright (C) 2008-2009  Wolfgang Ullrich
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/wait.h>
#include <syslog.h>
#include <malloc.h>
#include <errno.h>
#include <pwd.h>
#include <string>

#include "../include/globals.h"
#include "UserSettings.h"

using namespace std;

#define PASS_PROMPT  "Password: "
#define LOGIN_PROMPT "Login: "

#define PAM_SM_AUTH
#include <security/pam_modules.h>
#include <security/_pam_macros.h>

static string syslogIdent=string("fingerprintPAM");
static const char *pam_set_data_id="fingerprintPAM_return";

bool file_exists(const char* filename){
    if(FILE* file=fopen(filename,"r")){
        fclose(file);
        return true;
    }
    return false;
}

static void pam_set_data_free (pam_handle_t*,void *ptr,int){
    if(ptr)
        free(ptr);
}

void reopenSyslog(bool debug){  // Some evil stuff might have opened their own syslog meanwhile so we open ours again
    closelog();
    openlog(syslogIdent.data(),LOG_NDELAY|LOG_PID,LOG_AUTH);
    if(debug)
        setlogmask(-1);
    else
        setlogmask(LOG_UPTO(LOG_ERR));
}

// This function was derived from thinkfinger project (http://sourceforge.net/projects/thinkfinger)
// Executes the PAM conversation function of the calling application
int pam_vprompt(pam_handle_t *pamh,int style,char **response,const char *fmt,va_list args){
    struct pam_message msg;
    struct pam_response *pam_resp=NULL;
    const struct pam_message *pmsg;
    const struct pam_conv *conv;
    const void *convp;
    char *msgbuf;
    int ret_val;

    if(response) *response=NULL;
    ret_val=pam_get_item(pamh,PAM_CONV,&convp);
    if(ret_val!=PAM_SUCCESS)return ret_val;

    conv=(const struct pam_conv *)convp;
    if(conv==NULL||conv->conv==NULL){
        syslog(LOG_ERR,"No conversation function.");
        return PAM_SYSTEM_ERR;
    }

    if(vasprintf(&msgbuf,fmt,args)<0){
        syslog(LOG_ERR,"ERROR: Calling vasprintf.");
        return PAM_BUF_ERR;
    }

    msg.msg_style=style;
    msg.msg=msgbuf;
    pmsg=&msg;
    ret_val=conv->conv(1,&pmsg,&pam_resp,conv->appdata_ptr);
    if(ret_val!=PAM_SUCCESS&&pam_resp!=NULL)
        syslog(LOG_WARNING,"WARNING: Unexpected response from failed conversation function.");

    if(response){
        *response=pam_resp==NULL?NULL:pam_resp->resp;
    }else{
        if(pam_resp&&pam_resp->resp){
            _pam_overwrite(pam_resp->resp);
            _pam_drop(pam_resp->resp);
        }
    }
    _pam_overwrite(msgbuf);
    _pam_drop(pam_resp);
    free(msgbuf);
    if(ret_val!=PAM_SUCCESS)
        syslog(LOG_ERR,"Conversation failed.");

    return ret_val;
}

// This function was derived from thinkfinger project (http://sourceforge.net/projects/thinkfinger)
// Prompts for username/password
int pam_prompt(pam_handle_t *pamh,int style,char **response,const char *fmt, ...){
    va_list args;
    int ret_val;

    va_start(args,fmt);
    syslog(LOG_DEBUG,"Prompting \"%s\"",fmt);
    ret_val=pam_vprompt(pamh,style,response,fmt,args);
    va_end(args);
    return ret_val;
}

// Call from PAM ---------------------------------------------------------------
PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh,int/*flags*/,int argc,const char **argv){
    const char *rhost=NULL;
    const char *service=NULL;
    const char *pamtty=NULL;
    const char *sshconn=NULL;
    char *username=NULL;
    char *password=NULL;
    const char *passwdPath=NULL;
    pid_t child;
    int fifo[2];
    bool debug=false;
    char identifiedUser[256];
    int rc=PAM_AUTHINFO_UNAVAIL;
    int *ret_data=(int*)malloc(sizeof(int));  //return value to be given to pam_set_data
    memset(identifiedUser,0,sizeof(identifiedUser));
    const char* display=getenv("DISPLAY");
    char* xauthpath=getenv("XAUTHORITY");
    char X_lock[100];
    int xlock,i;
    char randomString[11];
    
    openlog(syslogIdent.data(),LOG_NDELAY|LOG_PID,LOG_AUTH);
    setlogmask(LOG_UPTO(LOG_ERR));
    for(int i=0;i<argc;i++){
        if((strcmp(argv[i],"debug")==0)
                | (strcmp(argv[i],"--debug")==0)
                | (strcmp(argv[i],"-d")==0)){
            setlogmask(-1);
            debug=true;
            syslog(LOG_INFO,"Got \"debug\" argument.");
        }
        if((strcmp(argv[i],"--passwd")==0)
                | (strcmp(argv[i],"-p")==0)){
            passwdPath=argv[++i];
            syslog(LOG_DEBUG,"Have passwd path %s.",passwdPath);
        }
    }

    // We don't request fingerprints for remote sessions
    pam_get_item(pamh,PAM_RHOST,(const void **)(const void*)&rhost);
    if (rhost!=NULL&&strlen(rhost)>0){  // remote login
        syslog(LOG_DEBUG,"Remote session detected.");
        goto auth_return;
    }
    else{
        syslog(LOG_DEBUG,"Got no PAM_RHOST.");
    }
    if((sshconn=getenv("SSH_CONNECTION"))){  // This is a remote session
        syslog(LOG_DEBUG,"Got SSH_CONNECTION %s, guess this is a remote session.",sshconn);
        goto auth_return;
    }
    else{
        syslog(LOG_DEBUG,"Got no SSH_CONNECTION.");
    }

    srand((unsigned)time(NULL));
    memset(randomString,0,sizeof(randomString));
    sprintf(randomString,"%ld",1000000000L+(rand()%1000000000L));
    
    // If we have a password already other PAM modules may handle it.
    pam_get_item(pamh,PAM_AUTHTOK,(const void **)(const void*)&password);
    if(password!=NULL){
        syslog(LOG_INFO,"Have a password. Should be handled by other PAM modules.");
        goto auth_return;
    }
    else{
        syslog(LOG_DEBUG,"Got no PAM_AUTHTOK.");
    }
    
    // Requested service; for debugging only
    pam_get_item(pamh,PAM_SERVICE,(const void **)(const void*)&service);
    if (service!=NULL&&strlen(service)>0){
        syslog(LOG_DEBUG,"PAM_SERVICE: %s.",(const char*)service);
    }
    else{
        syslog(LOG_ERR,"Got no PAM_SERVICE.");
    }
    
    // Requesting TTY; for debugging only
    pam_get_item(pamh,PAM_TTY,(const void **)(const void*)&pamtty);
    if (pamtty!=NULL&&strlen(pamtty)>0){
        syslog(LOG_DEBUG,"PAM_TTY: %s.",(const char*)pamtty);
        if(pamtty[0]==':'){     // Maybe it's a X-Display
            if(display==NULL){  // Sometimes there is no DISPLAY variable set in this case
                display=pamtty;
                syslog(LOG_WARNING,"Have no DISPLAY. Setting DISPLAY to %s.",display);
                setenv("DISPLAY",display,-1);
            }
        }
    }
    else{
        syslog(LOG_DEBUG,"Got no PAM_TTY.");
    }
    
    // If an username is available we try to authenticate by fingerprint
    // Otherwise we try to identify the user
    pam_get_item(pamh,PAM_USER,(const void **)(const void*)&username);
    if(username!=NULL){
        if(strlen(username)==0){    // Uups, got an empty username
            username=NULL;
        }
    }
    syslog(LOG_DEBUG,"Have PAM user %s.",username);

    // We create two processes here:
    // Parent will handle the login/auth by username/password at the PAM prompt
    // Child will request fingerprints
    // When fingerprint was recognized child will send username/"random string password" to the PAM prompt
    
    // Create a fifo for the parent process to send "random string password" to child
    // Child resends this "random string password" to the PAM prompt, whent fingerprint was recognized
    if(pipe(fifo)==-1){
        syslog(LOG_ERR,"Creating IPC pipe failed.");
        goto auth_return;
    }
    
    if(display!=NULL&&xauthpath==NULL){  // We have a x-display, so we need XAUTHORITY to connect to it from child
        sprintf(X_lock,"/tmp/.X%s-lock",strtok((char*)&display[1],"."));
        syslog(LOG_DEBUG,"Scanning X-server PID from %s.",X_lock);
        xlock=open(X_lock,O_RDONLY);
        if(xlock<0){
            syslog(LOG_DEBUG,"ERROR: Open %s (%s).",X_lock,strerror(errno));
            goto auth_return;
        }
        if(read(xlock,X_lock,sizeof(X_lock))<=0){
            syslog(LOG_DEBUG,"ERROR: Read %s (%s).",X_lock,strerror(errno));
            close(xlock);
            goto auth_return;
        }
        close(xlock);
        sscanf(X_lock,"%d",&xlock);
        syslog(LOG_DEBUG,"Have PID %d.",xlock);
        sprintf(X_lock,"/proc/%d/cmdline",xlock);
        syslog(LOG_DEBUG,"Searching for \"-auth\" in %s.",X_lock);
        xlock=open(X_lock,O_RDONLY);
        if(xlock<0){
            syslog(LOG_DEBUG,"ERROR: Open %s (%s).",X_lock,strerror(errno));
            goto auth_return;
        }
        if((i=read(xlock,X_lock,sizeof(X_lock)))<=0){
            syslog(LOG_DEBUG,"ERROR: Read %s (%s).",X_lock,strerror(errno));
            close(xlock);
            goto auth_return;
        }
        close(xlock);
        for(int j=0;j<i;j++)if(X_lock[j]=='\0')X_lock[j]=' ';
        syslog(LOG_DEBUG,"Have cmdline %s.",X_lock);
        for(char *word=strtok(X_lock," ");word!=NULL;word=strtok(NULL," ")){
            if(strcmp(word,"-auth")==0){
                xauthpath=strtok(NULL," ");
                break;
            }
        }
        if(file_exists(xauthpath)){
            setenv("XAUTHORITY",xauthpath,-1);
            syslog(LOG_DEBUG,"Setting XAUTHORITY to %s.",xauthpath);
        }
    }
    
    syslog(LOG_DEBUG,"Have display=%s, XAUTHORITY=%s.",display,xauthpath);
    
    syslog(LOG_DEBUG,"Parent PID: %d.",getpid());
    child=fork();           // here we start a child process that requests fingerprints
    switch(child){
        case 0:             // This is the child
            close(fifo[1]); // Child only reads from fifo a random string used as "password"
            char pipestr[10];// Convert pipe handle to string to be given in argv
            sprintf(pipestr,"%d",fifo[0]);
            if(display!=NULL)
                sleep(1);// We want the helper window to be raised as last one for being visible on top
            // Because of some applications (like gnome-screensaver) behave nasty and crash
            // a QApplicaton when started in their context for an unknown reason,
            // we need to use "execl" for running the QApplication (PamGUI, PAMNonGUI)
            rc=execl(HELPER_COMMAND,"fingerprintHelper",pipestr,
                    debug?"-d":"dUmMy1",
                    username!=NULL?"-u":"dUmMy2",
                    username!=NULL?username:"dUmMy3",
                    passwdPath!=NULL?"-p":"dUmMy4",
                    passwdPath!=NULL?passwdPath:"dUmMy5",
                    NULL);
            reopenSyslog(debug);
            syslog(LOG_ERR,"ERROR: Child returned %d (%s).",rc,strerror(errno));
            _exit(EXIT_FAILURE);
        case -1:            // Fork error
            syslog(LOG_ERR,"ERROR FORKING CHILD PROCESS.");
            rc=-1;
            break;
        default:            // This is the parent
            close(fifo[0]); // Parent only writes to fifo what random string to be used as "password"
            syslog(LOG_DEBUG,"Child PID: %d.",child);
            if(write(fifo[1],randomString,strlen(randomString))<0)
                syslog(LOG_ERR,"ERROR: Writing random string to fifo.");
            else
                syslog(LOG_DEBUG,"Wrote random string to fifo.");
            if(username==NULL){        // user still unknown, prompt for username
                if(pam_prompt(pamh,PAM_PROMPT_ECHO_ON, &username,LOGIN_PROMPT)!=PAM_SUCCESS){
                    reopenSyslog(debug);
                    syslog(LOG_ERR,"ERROR: Prompting for username.");
                    //RETURN PAM_AUTHINFO_UNAVAIL and kill child
                    kill(child,SIGUSR2); // Stop child immediately
                    waitpid(child,&rc,0);
                    rc=PAM_AUTHINFO_UNAVAIL;  // Let others handle that
                    goto auth_return;
                }
                reopenSyslog(debug);
                if(strlen(username)==0){    // We have an empty username. Maybe the user has pressed <enter> only
                    syslog(LOG_WARNING,"WARNING: Got empty username.");
                    //RETURN PAM_AUTHINFO_UNAVAIL and kill child
                    kill(child,SIGUSR2); // Stop child immediately
                    waitpid(child,&rc,0);
                    rc=PAM_AUTHINFO_UNAVAIL;  // Let others handle that
                    goto auth_return;
                }                
                pam_set_item(pamh,PAM_USER,username);
                syslog(LOG_DEBUG,"Got username: %s.",username);
                // We send a SIGUSR1 to the child.
                // In case the user has typed the username the child should exit now
                syslog(LOG_DEBUG,"Waiting for helper to be ready for SIGUSR1.");
                usleep(100000);
                kill(child,SIGUSR1); // Stop child if it didn't send the username
            }
            // At this point we have a username. Let's prompt for a password
            if(pam_prompt(pamh,PAM_PROMPT_ECHO_OFF,&password,PASS_PROMPT)!=PAM_SUCCESS){
                reopenSyslog(debug);
                syslog(LOG_ERR,"ERROR: Prompting for password.");
                //RETURN PAM_AUTHINFO_UNAVAIL and kill child
                kill(child,SIGUSR2); // Stop child immediately
                waitpid(child,&rc,0);
                rc=PAM_AUTHINFO_UNAVAIL;  // Let others handle that
                goto auth_return;
            }
            reopenSyslog(debug);
//!!!!!!!!!!!!!!!! USE THIS LOG OUTPUT ONLY FOR DEVELOPMENT !!!!!!!!!!!!!!!!!!!!
//syslog(LOG_DEBUG,"Got password \"%s\" (%s).",password,randomString);
//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!            
            if(strcmp(randomString,password)==0){   // The correct "password" was given by fingerprintHelper
                //RETURN PAM_SUCCESS (Child exits itself)
                waitpid(child,&rc,0);
// Try to get a saved password from external media (USB stick)
                char *savedPassword;
                // Find user's homedir
                char homedir[FILENAME_MAX];
                struct passwd *passwd;
                passwd=getpwnam(username);
                if(!passwd){
                    syslog(LOG_ERR,"No pwd entry for user %s.",username);
                    goto auth_return;
                }
                sprintf(homedir,"%s",passwd->pw_dir);
                // Create a "UserSettings" object
                UserSettings userSettings(homedir);
                if(userSettings.getPassword(&savedPassword)){
//!!!!!!!!!!!!!!!! USE THIS LOG OUTPUT ONLY FOR DEVELOPMENT !!!!!!!!!!!!!!!!!!!!
//syslog(LOG_INFO,"Have password %s from external media.",savedPassword);
//!!!!!!!!!!!!!!!! USE THIS LOG OUTPUT ONLY FOR DEVELOPMENT !!!!!!!!!!!!!!!!!!!!
                    if(pam_set_item(pamh,PAM_AUTHTOK,savedPassword)==PAM_SUCCESS){
                        syslog(LOG_DEBUG,"Success setting password from external media.");
                    }
                    free(savedPassword);
                    rc=PAM_IGNORE;  // Let others check the given username/password
                    goto auth_return;
                }
                // Here we didn't find the password from external media, but have a successful fingerprint
                rc=PAM_SUCCESS;
                goto auth_return;
            }
            // A "real" password was given by the user
            pam_set_item(pamh,PAM_AUTHTOK,password);
            kill(child,SIGUSR2); // Stop child immediately
            waitpid(child,&rc,0);
            syslog(LOG_DEBUG,"Got password by keyboard.");
            rc=PAM_IGNORE;  // Let others check the given username/password
            goto auth_return;
    }
    syslog(LOG_ERR,"This should not happen!");
    rc=PAM_AUTHINFO_UNAVAIL;  // Something went wrong
    
auth_return:
    *ret_data=rc;
    pam_set_data(pamh,pam_set_data_id,(void*)ret_data,pam_set_data_free);
    syslog(LOG_DEBUG,"Return %d (%s).",rc,
        rc==PAM_SUCCESS?"PAM_SUCCESS":
        rc==PAM_AUTHINFO_UNAVAIL?"PAM_AUTHINFO_UNAVAIL":
        rc==PAM_USER_UNKNOWN?"PAM_USER_UNKNOWN":
        rc==PAM_IGNORE?"PAM_IGNORE":
        rc==PAM_SYSTEM_ERR?"PAM_SYSTEM_ERR":
        rc==PAM_BUF_ERR?"PAM_BUF_ERR":
        "unknown");
    int pluginFifo=open(PLUGIN_FIFO,O_WRONLY|O_NDELAY);   // In case we have a fingerprintPlugin running, stop it
    if(pluginFifo<0){ // Maybe there is no fingerprintPlugin listening
        syslog(LOG_DEBUG,"Could not open named pipe (fingerprintPlugin not present).");
    }
    else{
        syslog(LOG_DEBUG,"Opened fifo to fingerprintPlugin: %d",pluginFifo);
        string fifoMsg(MSG_STOP); // Compose a message for fingerprintPlugin
        if(write(pluginFifo,fifoMsg.data(),strlen(fifoMsg.data())+1)<0)
            syslog(LOG_ERR,"ERROR: Writing to fifo: %s",strerror(errno));
        else
            syslog(LOG_DEBUG,"Wrote to fifo: %s",fifoMsg.data());
    }
    closelog();
    unlink(HELPER_PID); // In case there was a pidfile left over
    return rc;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh,int/*flags*/,int argc,const char **argv){
    int rc=PAM_SUCCESS;
    const void *prc=NULL;
    
    openlog(syslogIdent.data(),LOG_NDELAY|LOG_PID,LOG_AUTH);
    for(int i=0;i<argc;i++){
        if((strcmp(argv[i],"debug")==0)
                | (strcmp(argv[i],"--debug")==0)
                | (strcmp(argv[i],"-d")==0)){
            setlogmask(-1);
            syslog(LOG_INFO,"Got \"debug\" argument.");
        }
    }
    syslog(LOG_INFO,"Pam_sm_setcred was called.");
    if(pam_get_data(pamh,"fingerprintPAM_return",&prc)==PAM_SUCCESS&&prc){
        rc=*(const int *)prc;
        pam_set_data(pamh,"fingerprintPAM_return",NULL,NULL);
        syslog(LOG_INFO,"Returning (recovered) %d.",rc);
    }
    closelog();
    return rc;
}

// Don't know what this is good for...
// ...was never called so far...
// ...don't like to change auth token...
// ...but all the others have it, so me too ;-)
// BTW: Nothing written about PAM is worth be be called "documentation".
// It's a classical example for what a documentation never should be: written by
// the developer of a system.
PAM_EXTERN int pam_sm_chauthtok(pam_handle_t *pamh,int/*flags*/,int argc,const char **argv){
    int rc=PAM_SUCCESS;
    const void *prc=NULL;
    
    openlog(syslogIdent.data(),LOG_NDELAY|LOG_PID,LOG_AUTH);
    for(int i=0;i<argc;i++){
        if((strcmp(argv[i],"debug")==0)
                | (strcmp(argv[i],"--debug")==0)
                | (strcmp(argv[i],"-d")==0)){
            setlogmask(-1);
            syslog(LOG_INFO,"Got \"debug\" argument.");
        }
    }
    syslog(LOG_INFO,"Pam_sm_chauthtok was called.");
    if(pam_get_data(pamh,"fingerprintPAM_return",&prc)==PAM_SUCCESS&&prc){
        rc=*(const int *)prc;
        pam_set_data(pamh,"fingerprintPAM_return",NULL,NULL);
        syslog(LOG_INFO,"Returning (recovered) %d.",rc);
    }
    closelog();
    return rc;
}

#ifdef PAM_STATIC
struct pam_module _pam_thinkfinger_modstruct = {
	"fingerprintPAM",
	pam_sm_authenticate,
	pam_sm_setcred,
	NULL,
	NULL,
	NULL,
	pam_sm_chauthtok
};
#endif
